from collections import OrderedDict
from typing import Callable, List, Optional, Set, Tuple, cast

from mypy import message_registry
from mypy.checkexpr import ExpressionChecker
from mypy.exprtotype import expr_to_unanalyzed_type
from mypy.nodes import CallExpr, Context, DictExpr, Expression, GDEF, NameExpr, StrExpr, SymbolTableNode
from mypy.plugin import DynamicClassDefContext, Plugin
from mypy.semanal import SemanticAnalyzerPass2
from mypy.semanal_typeddict import TypedDictAnalyzer
from mypy.types import AnyType, Instance, Type, TypeOfAny, TypedDictType


class UnsupportedKeyTypeError(ValueError):
    pass


def parse_key_typeddict_fields(attrs_expr: DictExpr) -> Tuple[List[str], List[Type], Set[str]]:
    fields = []
    types = []
    required_fields = set()
    for field_name_expr, field_type_expr in attrs_expr.items:
        if isinstance(field_name_expr, StrExpr):
            fields.append(field_name_expr.value)
            required_fields.add(field_name_expr.value)

        elif isinstance(field_name_expr, CallExpr):
            fields.append(field_name_expr.args[0].value)
            required_expr = field_name_expr.args[1] if len(field_name_expr.args) == 2 else NameExpr('builtins.True')
            if required_expr.fullname == 'builtins.True':
                required_fields.add(field_name_expr.args[0].value)

        else:
            raise UnsupportedKeyTypeError(str(type(field_name_expr)))
        types.append(expr_to_unanalyzed_type(field_type_expr))

    return fields, types, required_fields


def check_typeddict_call_with_kwargs(self,
                                     callee: TypedDictType,
                                     kwargs: 'OrderedDict[str, Expression]',
                                     context: Context) -> Type:
    if not getattr(callee, 'allow_extra', False):
        if not (callee.required_keys <= set(kwargs.keys()) <= set(callee.items.keys())):
            expected_keys = [key for key in callee.items.keys()
                             if key in callee.required_keys or key in kwargs.keys()]
            actual_keys = kwargs.keys()
            self.msg.unexpected_typeddict_keys(
                callee,
                expected_keys=expected_keys,
                actual_keys=list(actual_keys),
                context=context)
            return AnyType(TypeOfAny.from_error)

    for (item_name, item_expected_type) in callee.items.items():
        if item_name in kwargs:
            item_value = kwargs[item_name]
            self.chk.check_simple_assignment(
                lvalue_type=item_expected_type, rvalue=item_value, context=item_value,
                msg=message_registry.INCOMPATIBLE_TYPES,
                lvalue_name='TypedDict item "{}"'.format(item_name),
                rvalue_name='expression')

    return callee


def copy_modified(self, *, fallback: Optional[Instance] = None,
                  item_types: Optional[List[Type]] = None,
                  required_keys: Optional[Set[str]] = None) -> 'TypedDictType':
    if fallback is None:
        fallback = self.fallback
    if item_types is None:
        items = self.items
    else:
        items = OrderedDict(zip(self.items, item_types))
    if required_keys is None:
        required_keys = self.required_keys

    new_type = TypedDictType(items, required_keys, fallback, self.line, self.column)
    new_type.allow_extra = getattr(self, 'allow_extra', False)
    return new_type


ExpressionChecker.check_typeddict_call_with_kwargs = check_typeddict_call_with_kwargs
TypedDictType.copy_modified = copy_modified


def add_key_typeddict_to_global_symboltable(ctx: DynamicClassDefContext) -> None:
    api = cast(SemanticAnalyzerPass2, ctx.api)
    typeddict_analyzer = TypedDictAnalyzer(options=api.options,
                                           api=api,
                                           msg=api.msg)
    try:
        fields, types, required_fields = parse_key_typeddict_fields(ctx.call.args[1])
    except UnsupportedKeyTypeError as error:
        api.fail(f'Unsupported key type {error.args[0]}', ctx=ctx.call)

    info = typeddict_analyzer.build_typeddict_typeinfo(ctx.name,
                                                       items=fields,
                                                       types=types,
                                                       required_keys=required_fields)
    allow_extra = api.parse_bool(ctx.call.args[2]) if len(ctx.call.args) > 2 else False
    info.typeddict_type.allow_extra = allow_extra
    api.add_symbol_table_node(ctx.name,
                              SymbolTableNode(GDEF, info))


class KeyTypedDictPlugin(Plugin):
    def get_dynamic_class_hook(self, fullname: str
                               ) -> Optional[Callable[[DynamicClassDefContext], None]]:
        if fullname == 'key_typeddict.core.KeyTypedDict':
            return add_key_typeddict_to_global_symboltable
        return None


def plugin(version):
    return KeyTypedDictPlugin



import dataclasses
import sys
import typing

from mypy_extensions import _check_fails, _dict_new  # type: ignore


@dataclasses.dataclass(eq=True, frozen=True)
class Key(object):
    name: str
    required: bool = True


def _key_typeddict_new(cls, _typename, _fields=None, **kwargs):
    total = kwargs.pop('total', True)
    if _fields is None:
        _fields = kwargs
    elif kwargs:
        raise TypeError("TypedDict takes either a dict or keyword arguments,"
                        " but not both")
    return _KeyTypedDictMeta(_typename, (), {'__annotations__': dict(_fields),
                                                  '__total__': total})


class _KeyTypedDictMeta(type):
    def __new__(cls, name, bases, ns, total=True, allow_extra=False):
        # Create new typed dict class object.
        # This method is called directly when TypedDict is subclassed,
        # or via _typeddict_new when TypedDict is instantiated. This way
        # TypedDict supports all three syntaxes described in its docstring.
        # Subclasses and instances of TypedDict return actual dictionaries
        # via _dict_new.
        ns['__new__'] = _key_typeddict_new if name == 'KeyTypedDict' else _dict_new
        tp_dict = super(_KeyTypedDictMeta, cls).__new__(cls, name, (dict,), ns)
        try:
            # Setting correct module is necessary to make typed dict classes pickleable.
            tp_dict.__module__ = sys._getframe(2).f_globals.get('__name__', '__main__')
        except (AttributeError, ValueError):
            pass
        anns = ns.get('__annotations__', {})
        msg = "KeyTypedDict('Name', {f0: t0, f1: t1, ...}); each t must be a type"
        anns = {n: typing._type_check(tp, msg) for n, tp in anns.items()}
        for base in bases:
            anns.update(base.__dict__.get('__annotations__', {}))
        tp_dict.__annotations__ = anns
        if not hasattr(tp_dict, '__total__'):
            tp_dict.__total__ = total
        if not hasattr(tp_dict, '__allow_extra__'):
            tp_dict.__allow_extra__ = allow_extra
        return tp_dict

    __instancecheck__ = __subclasscheck__ = _check_fails


KeyTypedDict = _KeyTypedDictMeta('KeyTypedDict', (dict,), {})
KeyTypedDict.__module__ = __name__



from typing import Any, Dict, Type, Union

# mypy_extensions.TypedDict should be present in the dependency graph
# noinspection PyUnresolvedReferences
from mypy_extensions import TypedDict


class Key:
    def __init__(self, name: str, required: bool = False): ...


def KeyTypedDict(typename: str,
                      fields: Dict[Union[str, Key], Type[Any]],
                      *,
                      total: bool = ...,
                      allow_extra: bool = ...) -> Type[dict]: ...



import sys
from distutils.core import setup

from setuptools import find_packages


dependencies = [
    'mypy>=0.670'
]
if sys.version_info[:2] < (3, 7):
    # dataclasses port for 3.6
    dependencies += ['dataclasses']

setup(
    name="key-typeddict",
    version="0.1.0",
    description='mypy_extensions.TypedDict with some additional features',
    url="https://github.com/mkurnikov/key-typeddict",
    author="Maksim Kurnikov",
    author_email="maxim.kurnikov@gmail.com",
    license='MIT',
    install_requires=dependencies,
    packages=['key_typeddict'],
    package_data={'key_typeddict': ['py.typed', 'core.pyi']},
    zip_safe=False
)
